# A class to use PyGame window as a "display"

import sys
import time
from PIL import Image
import pygame
import displayio

from dataclasses import astuple

# Our pygame "display" doesn't need any init sequence
_INIT_SEQUENCE = None

class PygameDisplay(displayio.Display):

    def __init__(self, _width, _height, **kwargs):
        self.running = True
        self._width = _width
        self._height = _height
        # initialize the super class, displayio.Display.
        #   DisplayBus = None
        #   
        super().__init__(None, _INIT_SEQUENCE, width=_width, height=_height, **kwargs)

    def _initialize(self, init_sequence):
        pygame.init()
        pygame.display.set_caption("Shulltronics Displayio PyGame Testing")
        self._pygame_screen = pygame.display.set_mode((self._width, self._height))
    
    # def _refresh(self):
    #     for event in pygame.event.get():
    #         if event.type == pygame.QUIT:
    #             sys.exit()


    def refresh(self, *, target_frames_per_second=60, minimum_frames_per_second=1):
        """
        When auto refresh is off, waits for the target frame rate and then refreshes the
        display, returning True. If the call has taken too long since the last refresh call
        for the given target frame rate, then the refresh returns False immediately without
        updating the screen to hopefully help getting caught up.
        If the time since the last successful refresh is below the minimum frame rate, then
        an exception will be raised. Set minimum_frames_per_second to 0 to disable.
        When auto refresh is on, updates the display immediately. (The display will also
        update without calls to this.)
        """

        if self.running:
            self._subrectangles = []

            # Go through groups and and add each to buffer
            if self._core._current_group is not None:

                buffer = Image.new("RGBA", (self._core._width, self._core._height))
                # Recursively have everything draw to the image
                # pylint: disable=protected-access
                self._core._current_group._fill_area(
                    buffer
                )  # pylint: disable=protected-access
                # save image to buffer (or probably refresh buffer so we can compare)
                self._buffer.paste(buffer)

            self._subrectangles = self._core.get_refresh_areas()

            for area in self._subrectangles:
                self._refresh_display_area(area)


    def _refresh_display_area(self, rectangle):
        """Loop through dirty rectangles and redraw that area."""

        img = self._buffer.convert("RGB").crop(astuple(rectangle))
        img = img.rotate(self._rotation, expand=True)
        display_rectangle = self._apply_rotation(rectangle)
        img = img.crop(astuple(self._clip(display_rectangle)))
        raw_str = img.tobytes("raw", "RGB")
        pygame_surface = pygame.image.fromstring(
            raw_str, (img.width, img.height), "RGB"
        )
        # print("({}, {})".format(img.width, img.height))
        self._pygame_screen.blit(pygame_surface, (rectangle.x1, rectangle.y1))
        pygame.display.flip()

    def get_mouse_clicks(self):

        event_return = None
        # pylint: disable=no-member, unused-argument)
        for event in pygame.event.get():
            # filter based on event:
            if event.type == pygame.QUIT:
                self.running = False
                time.sleep(0.1)
                pygame.quit()
            elif event.type == pygame.MOUSEBUTTONDOWN:
                event_return = pygame.mouse.get_pos()

        return event_return

    def quit(self):
        """
        Close the program!
        """
        print("Closing the program... goodbye!")
        self.running = False
        time.sleep(0.1)
        pygame.quit()


"""
A GUI framework built upon Adafruit displayio and the widget idea
This file should not use an platform-dependent code, so that
it can be run on any platform.
The user will define the display for their platform as follows:
GENERIC_PC (Windows or Linux), use PygameDisplay
Raspberry Pi or Microcontroller, use appropriate display_bus and Display
"""

from unigui.widget import *
from unigui.colorscheme import *
import displayio
from adafruit_bitmap_font import bitmap_font
from adafruit_display_text import label
from random import randint
import math


class UniGui(Widget):
    """
    This is the base class for my GUI
    An instance represents the root window with the following properties:
        * always located at 0, 0
        * size is defined upon creation, and can be scaled up by integer increments
        * has a colorscheme that is stored as a palette
        * TODO has a border that can be turned on or off
    """
    def __init__(self, width, height, colorscheme, scale=1):
        super().__init__(
            "UniGui",
            0,
            0,
            width,
            height,
            colorscheme,
            scale=scale,
        )
        self.widgets = []
        self.set_background()

    def add_widget(self, widget):
        # TODO: first validate the layout?

        # If we didn't override the widget's colorscheme, set it here
        if widget.palette is None:
            widget.palette = self.palette
        self.widgets.append(widget)
        self.append(widget)

    # returns the Widget named <name>, or None
    def get_widget(self, name):
        match = None
        for widget in self.widgets:
            if widget.name == name:
                match = widget
        return match

    # This method makes sure that the widgets don't overlap
    def validate_layout(self):
        for widget in self.widgets:
            pass  #TODO

    # Show the GUI on the display
    def update(self, display):
        display.show(self)

    # Process a click at display location (x, y)
    # We must account for screen scaling in this method
    def process_click(self, click_pos):
        (_x, _y) = click_pos
        x = round(_x / self.scale)
        y = round(_y / self.scale)
        print("(" + str(x) + ", " + str(y) + ")")
        for widget in self.widgets:
            widget_idx = self.index(widget)
            x_min = widget.x
            x_max = widget.x + widget.width
            y_min = widget.y
            y_max = widget.y + widget.height
            contains_x = (x >= x_min) and (x < x_max)
            contains_y = (y >= y_min) and (y < y_max)
            if contains_x and contains_y and widget.clickable:
                print("Widget index " + str(widget_idx) + " is a match! " + "Doing callback...")
                relative_click_pos = (x - widget.x, y - widget.y)
                widget.callback(relative_click_pos)



import displayio

"""
A ColorScheme is a palette of size 8
"""
class ColorScheme(displayio.Palette):
    indices = {
        'TRANSPARENT':    0,
        'BASE':           1,
        'BASE_HIGHLIGHT': 2,
        'TEXT':           3,
        'TEXT_HIGHLIGHT': 4,
        'COLOR_0':        5,
        'COLOR_1':        6,
        'COLOR_2':        7,
    }


class Shulltronics(ColorScheme):
    WHITE      = 0xFFFFFF
    BLACK      = 0x000000
    LIGHT_GRAY = 0x707070
    DARK_GRAY  = 0x454545
    RED        = 0xFF0000
    GREEN      = 0x00FF00
    BLUE       = 0x0000FF
    MAGENTA    = 0xFF00FF

    dark = displayio.Palette(8)
    dark[ColorScheme.indices['TRANSPARENT']]    = 0          # Transparent
    dark[ColorScheme.indices['BASE']]           = BLACK      # Background
    dark[ColorScheme.indices['BASE_HIGHLIGHT']] = DARK_GRAY  # Background highlights
    dark[ColorScheme.indices['TEXT']]           = LIGHT_GRAY # Default text
    dark[ColorScheme.indices['TEXT_HIGHLIGHT']] = WHITE      # Emphasized text
    dark[ColorScheme.indices['COLOR_0']]        = BLUE
    dark[ColorScheme.indices['COLOR_1']]        = RED
    dark[ColorScheme.indices['COLOR_2']]        = MAGENTA


class Solarized(ColorScheme):
    BASE03  = 0x002B36
    BASE02  = 0x073642
    BASE01  = 0x586E75
    BASE00  = 0x657B83
    BASE0   = 0x839496
    BASE1   = 0x93A1A1
    BASE2   = 0xEEE8D5
    BASE3   = 0xFDF6E3
    YELLOW  = 0xB58900
    ORANGE  = 0xCB4B16
    RED     = 0xDC322F
    MAGENTA = 0xD33682
    VIOLET  = 0x6C71C4
    BLUE    = 0x268BD2
    CYAN    = 0x2AA198
    GREEN   = 0x859900

    light = displayio.Palette(8)
    light[ColorScheme.indices['TRANSPARENT']]    = 0        # Transparent
    light[ColorScheme.indices['BASE']]           = BASE3    # Background
    light[ColorScheme.indices['BASE_HIGHLIGHT']] = BASE03   # Background highlights
    light[ColorScheme.indices['TEXT']]           = BASE00   # Default text
    light[ColorScheme.indices['TEXT_HIGHLIGHT']] = BASE01   # Emphasized text
    light[ColorScheme.indices['COLOR_0']]        = BLUE     # Highlight 1
    light[ColorScheme.indices['COLOR_1']]        = CYAN     # Highlight 2
    light[ColorScheme.indices['COLOR_2']]        = MAGENTA  # Highlight 3
    
    dark = displayio.Palette(8)
    dark[ColorScheme.indices['TRANSPARENT']]    = 0
    dark[ColorScheme.indices['BASE']]           = BASE03
    dark[ColorScheme.indices['BASE_HIGHLIGHT']] = BASE02
    dark[ColorScheme.indices['TEXT']]           = BASE2
    dark[ColorScheme.indices['TEXT_HIGHLIGHT']] = BASE3
    dark[ColorScheme.indices['COLOR_0']]        = BLUE
    dark[ColorScheme.indices['COLOR_1']]        = GREEN
    dark[ColorScheme.indices['COLOR_2']]        = RED


class VSCode():
    BASE03  = 0x000000
    BASE02  = 0x202020
    BASE01  = 0x586E75
    BASE00  = 0x657B83
    BASE0   = 0x839496
    BASE1   = 0x93A1A1
    BASE2   = 0xEEE8D5
    BASE3   = 0xFDF6E3
    YELLOW  = 0xB58900
    ORANGE  = 0xF38518
    RED     = 0xFF0000
    MAGENTA = 0xD33682
    VIOLET  = 0xB267E6
    BLUE    = 0x6796E6
    CYAN    = 0x9CDCFE
    GREEN   = 0x008000

    light = displayio.Palette(8)
    light[ColorScheme.indices['TRANSPARENT']]    = 0x010101        # Transparent
    light[ColorScheme.indices['BASE']]           = BASE3    # Background
    light[ColorScheme.indices['BASE_HIGHLIGHT']] = BASE03    # Background highlights
    light[ColorScheme.indices['TEXT']]           = ORANGE   # Default text
    light[ColorScheme.indices['TEXT_HIGHLIGHT']] = BASE01   # Emphasized text
    light[ColorScheme.indices['COLOR_0']]        = BLUE     # Highlight 1
    light[ColorScheme.indices['COLOR_1']]        = CYAN     # Highlight 2
    light[ColorScheme.indices['COLOR_2']]        = MAGENTA  # Highlight 3
    
    dark = displayio.Palette(8)
    dark[ColorScheme.indices['TRANSPARENT']]    = 0x010101
    dark[ColorScheme.indices['BASE']]           = BASE03
    dark[ColorScheme.indices['BASE_HIGHLIGHT']] = BASE02
    dark[ColorScheme.indices['TEXT']]           = ORANGE
    dark[ColorScheme.indices['TEXT_HIGHLIGHT']] = BASE03
    dark[ColorScheme.indices['COLOR_0']]        = BLUE
    dark[ColorScheme.indices['COLOR_1']]        = GREEN
    dark[ColorScheme.indices['COLOR_2']]        = ORANGE



from unigui.colorscheme import *
import displayio
import terminalio
from adafruit_bitmap_font import bitmap_font
from adafruit_display_text import label
import adafruit_imageload
from adafruit_display_shapes.rect import Rect
from adafruit_display_shapes.triangle import Triangle
from adafruit_display_shapes.circle import Circle
import math
from random import randint
# TODO: See if I can get this functionality working with CircuitPython
# from pkg_resources import resource_filename

class Widget(displayio.Group):
    """
    A Widget has a name, (x, y) location (upper left corner),
    and a (w, h) size in pixels.
    A Widget is a displayio.Group, and thus has the following properties:
    - It can have a list of children
    - etc..
    It also has the following properties:
    - it can be attached to events and be updated
    - it will adopt the GUI's colorshceme unless an overide colorscheme is passed as an argument
    """
    def __init__(self, name, x, y, width, height, colorscheme, scale=1):
        super().__init__(scale=scale)
        self.name          = name
        self.palette       = colorscheme
        self.palette.make_transparent(0)
        self.bg_transparent = False
        self.x             = x
        self.y             = y
        self.width         = width
        self.height        = height
        self.border        = False
        self.background_bm = None
        self.background_tg = None
        # by default a widget doesn't do anything when clicked
        self.clickable     = False
        # the callback function should take the click location tuple (xpos, ypos) as an argument
        self.callback      = None
        self.set_background()

    def get_background_tilegrid(self):
        """
        Return the tuple (index, TileGrid) that represents the widget's background.
        If the TileGrid already exists in the widget, it pops and returns that one,
        otherwise it creates and returns it without appending it to the group.
        """
        try:
            idx = self.index(self.background_tg)
            return (idx, self.pop(idx))
        except:
            print("Creating background tilegrid...")
            bg_bitmap = displayio.Bitmap(self.width, self.height, 8)
            self.background_bm = bg_bitmap
            self.background_tg = displayio.TileGrid(bg_bitmap, pixel_shader=self.palette)
            return (0, self.background_tg)

    def set_bg_transparent(self, trans):
        """ Sets the transparency of the widget's background (True or False). """
        self.bg_transparent = trans
        self.set_background()

    def set_background(self, bg_color_index=ColorScheme.indices['BASE']):
        """
        Set the background to the specified colorscheme index.
        """
        (tg_idx, tg) = self.get_background_tilegrid()   # Get the bg TileGrid
        if self.bg_transparent:
            self.background_bm.fill(0)
        else:
            self.background_bm.fill(bg_color_index)
        new_tg = displayio.TileGrid(self.background_bm, pixel_shader=self.palette)
        self.background_tg = new_tg
        self.insert(tg_idx, self.background_tg)

    def border_on(self, color_idx=ColorScheme.indices['BASE_HIGHLIGHT']):
        """
        Draw a 1px wide border around the widget.
        """
        self.border = True
        bm = self.background_bm
        # get the index of the background tilegrid.
        # Don't care about the tg itself because we're making a new one
        (tg_index, _) = self.get_background_tilegrid()
        for x in range(0, bm.width):
            bm[x, 0]           = color_idx
            bm[x, bm.height-1] = color_idx
        for y in range(0, bm.height):
            bm[0, y]          = color_idx
            bm[bm.width-1, y] = color_idx
        self.background_tg = displayio.TileGrid(bm, pixel_shader=self.palette)
        self.insert(tg_index, self.background_tg)

    def border_off(self):
        """
        Clear the border around the widget.
        """
        self.border = False
        bm = self.background_bm
        # get the index of the background tilegrid.
        # Don't care about the tg itself because we're making a new one
        (tg_index, _) = self.get_background_tilegrid()
        # Fill with the base color index
        bm.fill(ColorScheme.indices['BASE'])
        self.background_tg = displayio.TileGrid(bm, pixel_shader=self.palette)
        self.insert(tg_index, self.background_tg)

    def border_toggle(self, click_pos):
        if self.border:
            self.border_off()
        else:
            self.border_on()

    # Set the function to call when the widget is clicked
    def set_click_action(self, function):
        self.clickable = True
        self.callback = function

class TextWidget(Widget):
    """
    A Widget that displays a string nicely
        TODO: Add padding 
        TODO: ability to wrap, cutoff, or scroll long text
    """

    # TODO: See if I can get this functionality working with CircuitPython
    # SMALL_FONT = resource_filename('unigui', 'fonts/6x12.bdf')
                 #"fonts/VCROSDMono-14.bdf"
    # LARGE_FONT = resource_filename('unigui', 'fonts/fipps-12pt.bdf')
                 #"fonts/SNES-Italic-24.bdf"
                 #"fonts/Silom-Bold-24.bdf"

    def __init__(self, name, x, y, width, height, colorscheme, font_path=None):
        super().__init__(name, x, y, width, height, colorscheme)
        self.value = ""
        if font_path:
            self.font = bitmap_font.load_font(font_path)
        else:
            self.font = terminalio.FONT
        self.color = self.palette[ColorScheme.indices['TEXT']]
        self.label = label.Label(self.font, text=self.value, background=None, color=self.color)
        self.append(self.label)

    def get_label(self):
        try:
            idx = self.index(self.label)
            return (idx, self.pop(idx))
        except:
            print("Tried to get non-existant text label!")
            return None

    def set_value(self, value, h_justification='center', v_justification='center'):
        """
        Set the text that the widget should display.
        """
        if self.value == value:
            return
        (label_idx, label) = self.get_label()
        if label:
            self.value = value
            # Set the text value and then recalculate its position
            label.text = self.value
            label.anchor_point = (0, 0)
            bb = label._bounding_box
            (w, h) = (bb[2] - bb[0], bb[3] - bb[1])
            #TODO make this an object parameter
            x_padding = y_padding = 3

            if v_justification == 'top':
                y = y_padding
            elif v_justification == 'center':
                y = round(self.height/2) - round(h/2)
            elif v_justification == 'bottom':
                y = self.height - h
            
            if h_justification == 'left':
                x = x_padding
            elif h_justification == 'center':
                x = round(self.width/2) - round(w/2)
            elif h_justification == 'right':
                x = self.width - w - x_padding
            
            label.anchored_position = (x, y)
            self.insert(label_idx, label)

    # sets the text color
    def set_color(self, color_idx=ColorScheme.indices['TEXT']):
        """
        Set the text color using an index into the colorscheme
        """
        self.color = self.palette[color_idx]
        self.label.color = self.color

class IconWidget(Widget):

    # TODO: See if I can get this functionality working with CircuitPython
    # ICON_BUILTIN = resource_filename('unigui', 'images/128x32/px_icons.bmp')

    def __init__(self, name, x, y, width, height, colorscheme):
        self.icon_path = self.ICON_BUILTIN
        self.bm, self.icon_palette = adafruit_imageload.load(
            self.icon_path,
            bitmap=displayio.Bitmap,
            palette=displayio.Palette,
        )
        # print("icon palette length is ", self.icon_palette.__len__())
        super().__init__(name, x, y, width, height, colorscheme=colorscheme)
        # Setup the icon palette to the color we want, making index 0 transparent
        # Be sure when creating icons to make the background color be index 0
        self.icon_palette.make_transparent(0)
        self.icon_palette[1] = self.palette[ColorScheme.indices['COLOR_2']]
        self.tg = displayio.TileGrid(
            self.bm,
            pixel_shader=self.icon_palette,
            width=1,
            height=1,
            tile_width=32,
            tile_height=32,
            default_tile=2,
        )
        self.append(self.tg)

    def set_icon_index(self, idx=0):
        self.tg[0] = idx

class GraphicsWidget(Widget):
    """
    This widget draws some graphics
    """
    def __init__(self, name, x, y, width, height, colorscheme):
        super().__init__(name, x, y, width, height, colorscheme)
        self.graphics = None

    def get_graphics_group(self):
        try:
            idx = self.index(self.graphics)
            return (idx, self.pop(idx))
        except:
            print("Tried to get non-existant graphics group!")
            return None

    # Draws a random triangle centered in the widget
    def set_main_area(self, click_pos):
        (offset_x, offset_y) = (round(self.width/2), round(self.height/2))
        padding = 10 # TODO: paramatrize this value
        r = min(offset_x, offset_y) - padding
        # Get three random angles around our circle:
        p0_angle = math.radians(randint(0, 360))
        p1_angle = math.radians(randint(0, 360))
        p2_angle = math.radians(randint(0, 360))
        (x0, y0) = (round(r*math.sin(p0_angle)) + offset_x, round(r*math.cos(p0_angle)) + offset_y)
        (x1, y1) = (round(r*math.sin(p1_angle)) + offset_x, round(r*math.cos(p1_angle)) + offset_y)
        (x2, y2) = (round(r*math.sin(p2_angle)) + offset_x, round(r*math.cos(p2_angle)) + offset_y)
        triangle = Triangle(x0, y0, x1, y1, x2, y2,
            fill=self.palette[ColorScheme.indices['COLOR_0']],
            outline=self.palette[ColorScheme.indices['COLOR_1']],
        )
        # if the graphics already exist, update them
        if self.graphics:
            (graphics_idx, graphics) = self.get_graphics_group()
            self.graphics = triangle
            self.insert(graphics_idx, self.graphics)
        else:
            self.graphics = triangle
            self.append(self.graphics)
